@@ -3,7 +3,7 @@ use strict;
 use warnings;
 use base 'Text::MicroTemplate::File';
 
-our $VERSION = '0.12';
+our $VERSION = '0.13';
 
 sub new {
     my $self = shift->SUPER::new(@_);
@@ -107,6 +107,9 @@ sub render_file {
         local $@;
         eval {
             $result = $renderer->(@_);
+
+            my $tmpl = $self->{template};
+            $_->{template_ref} ||= \$tmpl for values %{ $context->{blocks} };
         };
         $die_msg = $@;
     }
@@ -132,7 +135,17 @@ sub _render_block {
     local ${"$self->{package_name}::_MTEREF"} = $block_ref;
 
     $$block_ref = '';
-    my $result = $block->{code}->() || $$block_ref || '';
+    my ($result, $die_msg);
+    eval {
+        $result = $block->{code}->() || $$block_ref || '';
+    };
+    if ($@) {
+        my $context = $self->render_context;
+        local $self->{template} = ${ $block->{template_ref} };
+        die $self->_error($@, 2 + $context->{offset}, $context->{caller});
+    }
+
+    $result;
 }
 
 sub include_file {
@@ -163,6 +176,7 @@ sub build {
     }->();
 
     $context->{args} = '';
+    my $offset = 0;
     for my $key (keys %{ $self->template_args || {} }) {
         unless ($key =~ /^[a-zA-Z_][a-zA-Z0-9_]*$/) {
             die qq{Invalid template args key name: "$key"};
@@ -174,9 +188,11 @@ sub build {
         else {
             $context->{args} .= qq{my \$$key = \$self->template_args->{$key};\n};
         }
+        $offset++;
     }
 
     $context->{blocks} ||= {};
+    $context->{offset} = $offset;
 
     my $die_msg;
     {
@@ -184,24 +200,22 @@ sub build {
         if (my $builder = $self->eval_builder) {
             return $builder;
         }
-        $die_msg = $self->_error($@, 4, $context->{caller});
+        $die_msg = $self->_error($@, 2 + $offset, $context->{caller});
     }
     die $die_msg;
 }
 
 sub eval_builder {
-    my $self = shift;
+    my ($self, $offset) = @_;
 
     local $SIG{__WARN__} = sub {
-        print STDERR $self->_error(shift, 4, $self->render_context->{caller});
+        print STDERR $self->_error(shift, 2 + $offset, $self->render_context->{caller});
     };
 
     eval <<"...";
 package $self->{package_name};
 sub {
-#line 1
-    $self->{render_context}{args};
-    Text::MicroTemplate::encoded_string(($self->{render_context}{code})->(\@_));
+    $self->{render_context}{args}Text::MicroTemplate::encoded_string(($self->{render_context}{code})->(\@_));
 }
 ...
 }
